classdef cuadrica
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%        

Esta 'clase' de Matlab permite definir una cuádrica bien con los
coeficientes de la ecuación coefs, bien con su matriz en coordenadas ampliada M.

Si se introduce M checkea si M es simétrica (ignora la parte
por debajo de la diagonal)

Está basada en el código de la clase "conica", y está mas limitada: solo 
proporciona la ecuacion reducida,
autovalores y autovectores, clasifica la cuadrica, y permite dibujarla.

INSTRUCCIONES

Para crear un nuevo objeto "cuadrica", basta con añadir al PATH de matlab (en el
menu FILE) la ruta a este archivo, y crearlo escribiendo en la consola de
matlab: 

    a = cuadrica

donde "a" es la variable que va a almacenar el objeto.

Para cambiar los coeficientes de la cuádrica, damos los nuevos coefs en forma 
de vector:

    a.coefs = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]; % ordenado x^2, y^2, z^2, xy, xz,
                                               % yz, x, y, z, 1

Si se pregunta cuanto vale "a", nos dará toda la información disponible
de la cuádrica:

    a    % Sin poner ";" al final, para que muestre por pantalla

Si queremos preguntar por un dato en concreto, por ejemplo, la matriz reducida,

    a.Mr % Sin ";" al final

Para dibujar la cuadrica con el resto de elementos, basta con invocar
"dibujar_todo" o sus variantes

    a.dibujar_todo;
    a.dibujar_todo_centrado;

Dudas, errores, sugerencias: mario.merino@upm.es

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%        
MMM20121212
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%        
%}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%        
    properties
        coefs = [1,4,8,0,0,0,0,0,0,-1]; % vector de coefs Ax^2 + By^2 + Cz^2 + Dxy + Exz + Fyz + Gx +Hy + Iz + J = 0
    end 
    properties (Dependent = true)
        texto_ecuacion; % mostrar la ecuacion inicial
        texto_ecuacion_r; % mostrar la ecuacion reducida
        x_; % la expresion del cambio de variable: reducidas en funcion de iniciales
        y_;
        z_;
        lambda_x_; % autovalores segun las direcciones x_, y_, z_
        lambda_y_;
        lambda_z_;
        autovector_x_; % autovectores segun esas direcciones
        autovector_y_;
        autovector_z_;
        tipo; % el tipo de cuadrica
        es_degenerada; % true/false si es degenerada
        es_reglada; % true/false si es reglada (existen familias de rectas en ella)
        M; % la matriz inicial, en coordenadas ampliadas
        Mr; % la matriz reducida, coordenadas ampliadas
        Qa; % Cambio de referencia desde lo inicial a lo reducido (c. ampliadas) X = Qa*X_
        ecuacion; % function handles (se usan internamente)
        ecuacion_r;
    end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    methods % metodos get/set
        function v = get.ecuacion(o)
            c = o.coefs;
            v =@(x,y,z) c(1)*x.^2 + c(2)*y.^2 + c(3)*z.^2 + ...
                        c(4)*x.*y + c(5)*x.*z + c(6)*y.*z + ...
                        c(7)*x + c(8)*y + c(9)*z + c(10);
        end
        function v = get.ecuacion_r(o)
            c = [o.Mr(2,2),o.Mr(3,3),o.Mr(4,4),...
                 2*o.Mr(2,3),2*o.Mr(2,4),2*o.Mr(3,4),...
                 2*o.Mr(1,2),2*o.Mr(1,3),2*o.Mr(1,4),o.Mr(1,1)];
            v =@(x_,y_,z_) c(1)*x_.^2 + c(2)*y_.^2 + c(3)*z_.^2 + ...
                        c(4)*x_.*y_ + c(5)*x_.*z_ + c(6)*y_.*z_ + ...
                        c(7)*x_ +c(8)*y_ + c(9)*z_ + c(10);
        end
        function v = get.texto_ecuacion(o)
            vars = {'x^2','y^2','z^2','x*y','x*z','y*z','x','y','z'};
            v = [o.expresion(o.coefs,vars),' = 0'];
        end
        function v = get.texto_ecuacion_r(o)
            c = [o.Mr(2,2),o.Mr(3,3),o.Mr(4,4),...
                 2*o.Mr(2,3),2*o.Mr(2,4),2*o.Mr(3,4),...
                 2*o.Mr(1,2),2*o.Mr(1,3),2*o.Mr(1,4),o.Mr(1,1)];
            vars = {'x_^2','y_^2','z_^2','x_*y_','x_*z_','y_*z_','x_','y_','z_'};
            v = [o.expresion(c,vars),' = 0'];
        end
        function v = get.x_(o)
            vars = {'x','y','z'};
            Qa_ = inv(o.Qa);
            v = o.expresion([Qa_(2,2:end),Qa_(2,1)],vars);
        end
        function v = get.y_(o)
            vars = {'x','y','z'};
            Qa_ = inv(o.Qa);
            v = o.expresion([Qa_(3,2:end),Qa_(3,1)],vars);
        end
        function v = get.z_(o)
            vars = {'x','y','z'};
            Qa_ = inv(o.Qa);
            v = o.expresion([Qa_(4,2:end),Qa_(4,1)],vars);
        end
        function v = get.lambda_x_(o)
            v = o.Mr(2,2);
        end
        function v = get.lambda_y_(o)
            v = o.Mr(3,3);
        end
        function v = get.lambda_z_(o)
            v = o.Mr(4,4);
        end
        function v = get.autovector_x_(o)
            v = o.Qa(2:end,2);
        end
        function v = get.autovector_y_(o)
            v = o.Qa(2:end,3);
        end
        function v = get.autovector_z_(o)
            v = o.Qa(2:end,4);
        end
        function v = get.M(o)
            % diagonal
            v(4,4) = o.coefs(3);
            v(3,3) = o.coefs(2);
            v(2,2) = o.coefs(1);
            v(1,1) = o.coefs(10);
            % fuera de la diagonal
            v(2,3) = o.coefs(4)/2; v(3,2) = v(2,3);
            v(2,4) = o.coefs(5)/2; v(4,2) = v(2,4);
            v(3,4) = o.coefs(6)/2; v(4,3) = v(3,4);
            v(1,2) = o.coefs(7)/2; v(2,1) = v(1,2);
            v(1,3) = o.coefs(8)/2; v(3,1) = v(1,3);
            v(1,4) = o.coefs(9)/2; v(4,1) = v(1,4);
        end
        function o = set.M(o,v)
            o.coefs = [v(2,2),v(3,3),v(4,4),2*v(2,3),2*v(2,4),2*v(3,4),2*v(1,2),2*v(1,3),2*v(1,4),v(1,1)];
        end
        function v = get.Mr(o)
            v = cuadrica.ecuacion_reducida(o.M);
        end
        function v = get.Qa(o)
            [~, v] = cuadrica.ecuacion_reducida(o.M);
        end
        function v = get.tipo(o) % esta es la función que clasifica la cuadrica segun su tipo
            sx = sign(o.lambda_x_);
            sy = sign(o.lambda_y_);
            sz = sign(o.lambda_z_);
            s1 = sign(o.Mr(1,1));            
            if sx*sy*sz ~= 0 % casos sin autovalores nulos
                if sx+sy+sz == 3 || sx+sy+sz == -3 % todos los autovalores igual signo (3,0) o (0,3)
                    if sx == -s1
                        v = 'elipsoide';
                        if o.lambda_x_ == o.lambda_y_ && o.lambda_x_ == o.lambda_z_
                            v = 'esfera'; % caso especial de elipsoide
                        end
                    elseif s1 == 0
                        v = 'un punto';
                    else
                        v = 'conjunto vacio (elipsoide complejo)';
                    end
                elseif sx+sy+sz == 1 || sx+sy+sz == -1 % signatura (2,1) o (1,2)
                    t = sx*sy*sz*s1;
                    if t == 1 % un unico autovalor tiene mismo signo que Mr(1,1)
                        v = 'hiperboloide 1 hoja';
                    elseif t == -1 % dos autovalores mismo signo que Mr(1,1)
                        v = 'hiperboloide 2 hojas';
                    else % s1 == 0
                        v = 'cono';
                    end
                end
            elseif sx*sy + sx*sz + sy*sz ~= 0 % un unico valor es cero
                t = sx*sy + sx*sz + sy*sz;
                if t == 1 % los dos que quedan, mismo signo
                    if sum(o.Mr(1,2:end)) ~= 0 
                        v = 'paraboloide eliptico';
                    elseif s1*sx + s1*sy + s1*sz == -2 % Mr(1,1) distinto signo
                        v = 'cilindro eliptico';
                    elseif s1*sx + s1*sy + s1*sz == 2 % Mr(1,1) igual signo
                        v = 'conjunto vacio (cilindro eliptico complejo)';
                    else % s1 == 0
                        v = 'una recta';
                    end
                elseif t == -1 % los dos que quedan, distinto signo
                    if sum(o.Mr(1,2:end)) ~= 0
                        v = 'paraboloide hiperbolico';
                    elseif s1 ~= 0 
                        v = 'cilindro hiperbolico';
                    else % s1 == 0
                        v = 'dos planos que se cortan';
                    end
                end
            elseif sx+sy+sz ~= 0 % un autovalor es no nulo
                if sum(o.Mr(1,2:end)) ~= 0
                    v = 'cilindro parabolico';
                elseif (sx+sy+sz)*s1 == -1
                    v = 'dos planos paralelos';                
                elseif (sx+sy+sz)*s1 == 1
                    v = 'conjunto vacio (dos planos complejos)';
                else % s1 == 0
                    v = 'un plano (doble)';
                end
            else % Los tres autovalores nulos
                if sum(o.Mr(1,2:end)) ~= 0 % hay parte lineal
                    v = 'un plano';                
                elseif s1 ~= 0
                    v = 'no hay solucion';
                end                    
            end
        end
        function v = get.es_degenerada(o)
            v = (det(o.M) == 0);
        end      
        function v = get.es_reglada(o) 
            v = any(strcmp(o.tipo,{'paraboloide hiperbolico','hiperboloide 1 hoja'}));
        end      
    end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    methods % metodos para dibujar
        function ph = dibujar_cuadrica(o,varargin)
            lims = 2*[sqrt(abs(o.Mr(1,1)/o.lambda_x_)),...
                      sqrt(abs(o.Mr(1,1)/o.lambda_y_)),...
                      sqrt(abs(o.Mr(1,1)/o.lambda_z_))];
            lims(isinf(lims)) = 2;
            lims(isnan(lims)) = 2;
            lims(lims==0) = 2;
            n_points = 60; % numero de puntos a utilizar para calcular isosurface
            [XX,YY,ZZ] = meshgrid(linspace(-lims(1),lims(1),n_points),...
                                  linspace(-lims(2),lims(2),n_points),...
                                  linspace(-lims(3),lims(3),n_points));
            FF = o.ecuacion_r(XX,YY,ZZ);
            % vuelve a referencia inicial
            Qa = o.Qa;            
            temp = Qa(2,2).*XX+Qa(2,3).*YY+Qa(2,4).*ZZ;
            temp2 = Qa(3,2).*XX+Qa(3,3).*YY+Qa(3,4).*ZZ;
            ZZ = Qa(4,2).*XX+Qa(4,3).*YY+Qa(4,4).*ZZ;
            XX = temp;
            YY = temp2;
            XX = XX+Qa(2,1);
            YY = YY+Qa(3,1);
            ZZ = ZZ+Qa(4,1);
            % dibujar
            set(gcf,'renderer','OpenGL');            
            [f,v,c] = isosurface(XX,YY,ZZ,FF,0,rand(size(XX)));
            ph = patch('Faces',f,'Vertices',v,...
                       'FaceVertexCData',c,...
                       'FaceColor','interp', ...
                       'edgecolor','none',...            
                       varargin{:});                   
            % transparencia segun el autovalor de signo distinto
            sx = sign(o.lambda_x_);
            sy = sign(o.lambda_y_);
            sz = sign(o.lambda_z_);
            if sx*sy*sz ~=0 && sx*sy + sx*sz + sy*sz == -1 % un autovalor distinto en signo
                if sx*sy == sx*sz % es x el distinto
                    maximo = max(2,2*sqrt(abs(o.Mr(1,1)/o.lambda_x_)));
                    vect = o.autovector_x_;
                elseif sy*sx == sy*sz % es y el distinto
                    maximo = max(2,2*sqrt(abs(o.Mr(1,1)/o.lambda_y_)));
                    vect = o.autovector_y_;
                else % es z
                    maximo = max(2,2*sqrt(abs(o.Mr(1,1)/o.lambda_z_)));
                    vect = o.autovector_z_;
                end
                alpha = v(:,1).*0;
                for i = 1:length(v(:,1))
                    alpha(i) = maximo - abs((v(i,:)-Qa(2:end,1)')*vect);
                end
                set(ph,'Facealpha','interp','edgealpha','interp',...                    
                       'FaceVertexAlphaData',sqrt(alpha),'Alphadatamapping','scaled');
            end
            set(gca,'CLim',[0,3],'ALim',[0,max(lims)/2]);
        end
        function ah = dibujar_todo(o)
            ah = gca;
            % centrar los ejes en la cuadrica:
            D = o.Qa(2:end,1); % el centro de la figura
            extra = 2*max(1,sqrt(abs(o.Mr(1,1))/max(abs([1,o.lambda_x_,o.lambda_y_,o.lambda_z_]))));
            set(ah,'XLim',D(1)+[-extra,+extra],'YLim',D(2)+[-extra,+extra],'ZLim',D(2)+[-extra,+extra]);  
            set(ah,'plotboxaspectratio',[1,1,1],'DataAspectRatio',[1,1,1]);
            % Dibujar todo
            dibujar_cuadrica(o);
            % Dar luz al dibujo
            camlight
            lighting gouraud
        end
        function ah = dibujar_todo_centrado(o)
            o2 = cuadrica;
            o2.M = o.Mr; % crea un nuevo objeto cuadrica y establece M = o.Mr 
            ah = o2.dibujar_todo;
        end
    end    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    methods (Static = true)
        function [Mr, Qa] = ecuacion_reducida(M)
            % Devuelve la forma reducida de la cuadrica a partir de una forma 
            % cualquiera dada en coordenadas ampliadas, Xt*M*X = 0
            %
            %     [  c   b1   b2   b3 ]        [  1 ]
            % M = [ b1  a11  a12  a13 ]    X = [ x1 ]
            %     [ b2  a12  a22  a23 ]        [ x2 ]
            %     [ b3  a13  a23  a33 ]        [ x3 ]
            %
            % Última revisión: MMM20121212
            
            % Submatrices A, B, C
            A = M(2:end,2:end);
            B = M(2:end,1);
            C = M(1,1);
            
            % Direcciones principales
            [Q,A] = eig(A);
            B = Q'*B; % cambia a ejes intermedios
            
            % Cambio de origen
            D = B.*0; % inicializa asi
            for i=1:length(A(1,:))
                if A(i,i) ~= 0 % cada autovalor
                    D(i) = -B(i)/A(i,i);
                    C = C + D(i)*B(i);
                    B(i) = 0;
                end
            end
            for i = 1:length(A(1,:))
                if A(i,i) == 0 && B(i) ~= 0
                    D(i) = -C/(2*B(i));
                    C = 0;
                end
            end            
            % Matriz de cambio de referencia
            Qa = [  1, D'*0;
                  Q*D,    Q];
            % Matriz reducida
            Mr = [C B';
                  B A];
            % Un ultimo paso: reducir B a un unico elemento no nulo
            indices_nucleo = [];
            vects_nucleo = zeros(3,0);
            indices_otros = [];
            vects_otros = zeros(3,0);
            for i = 1:length(A(1,:))
                if A(i,i) == 0
                    indices_nucleo(end+1) = i;
                    vects_nucleo(:,end+1) = Q(:,i);
                else
                    indices_otros(end+1) = i;
                    vects_otros(:,end+1) = Q(:,i);
                end
            end
            n = (B~=0);
            s = sum(n);
            if s > 1 % si hay mas de un elemento no nulo en B llegados a este punto 
                b = zeros(s);
                b(:,1) = B(n)/norm(B(n)); % matriz de coords en vectores del nucleo de los nuevos vectores                
                for j = 2:length(b);                    
                    t = - b(2:j,1:j-1)' \ b(1,1:j-1)';
                    b(1:j,j) = [1;t]/norm([1;t]);
                end
                Q_ = eye(length(B)); % prepara la matriz del giro (ampliada)
                Q_(n,n) = b;
                Qb(1,1) = 1;
                Qb(2:length(B)+1,2:length(B)+1) = Q_;
                Mr = Qb'*Mr*Qb; % aplica el cambio de ref. terciario
                Qa = Qa*Qb;
            end   
        end
        function eq = expresion(coefs,vars)
            % escribe el texto de una ecuacion dados los coefs y vars en orden.
            % coefs es un vector de dimension n+1, y vars un cellarray de
            % dimension n.
            % El ultimo coef es el termino indep, y no lleva vars asociada.
            % Si algun termino tiene coef 0, no lo escribe.
            % No se añade = 0 al final
            eq = '';
            for i=1:length(vars)
                if coefs(i) == 1
                    if isempty(eq) % inicialmente no poner '+'
                        eq = [eq, vars{i}];
                    else
                        eq = [eq,' + ', vars{i}];
                    end
                elseif coefs(i) == -1
                    eq = [eq,' - ', vars{i}];
                elseif coefs(i) > 0
                    if isempty(eq) % inicialmente no poner '+'
                        eq = [eq,num2str(coefs(i),3),'*',vars{i}];
                    else
                        eq = [eq,' + ',num2str(coefs(i),3),'*',vars{i}];
                    end
                elseif coefs(i) < 0
                    eq = [eq,' - ',num2str(abs(coefs(i)),3),'*',vars{i}];
                end
            end    
            if coefs(end) > 0
                if isempty(eq) % inicialmente no poner '+'
                    eq = [eq,num2str(coefs(end),3)];
                else
                    eq = [eq,' + ',num2str(coefs(end),3)];
                end
            elseif coefs(end) < 0
                eq = [eq,' - ',num2str(abs(coefs(end)),3)];
            end
        end
    end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
end

